/*
AntMap: Software for constructing genetic linkage maps using an Ant Colony Optimization (ACO) algorithm
Copyright (C) 2004 National Agricultural and Bio-oriented Research Organization (NARO)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

Hiroyoshi IWATA
Department of Information Science and Technology
National Agricultural Research Center
3-1-1 Kannondai
Tsukuba
Ibaraki 305-8666
JAPAN

ph        +81-298-38-7025
fax       +81-298-38-8551
e-mail    iwatah@affrc.go.jp
*/

/////////////////////////////////////////////////////////////////////////////////////
#include "ant.h"
/////////////////////////////////////////////////////////////////////////////////////

Ant::Ant(AcoParams *ap, Random *r) {
	params = ap;
	rand = r;
	n_city = params->getN_city();
	notVisited.resize(n_city);
	for(int i = 0; i < n_city; i++ ) {
		notVisited[i] = 1;
	}
	p.resize(n_city);
	tourOrder.resize(n_city);
	currentCity = 0;
};

void Ant::prepareForJourney() {
	for(int i = 0; i < n_city; i++ )
		notVisited[i] = 1;
	n_leg = 0;
}

void Ant::depart(double *choice_info) {
	if(params->isRandomize()) {
		if(rand->nextDouble() < params->getP_fancy())
			currentCity = selectStartPointRandomly();
		else
			currentCity = selectStartPoint(choice_info);
	} else {
		currentCity = selectStartPoint(choice_info);
	}
	tourOrder[n_leg] = currentCity;
	notVisited[currentCity] = 0;
	n_leg ++;
}

void Ant::move(double *choice_info) {
	if(rand->nextDouble() < 0.3) {
		currentCity = selectNextStopDetermin(choice_info);
	} else if(params->isRandomize()) {
		if(rand->nextDouble() < params->getP_fancy())
			currentCity = selectNextStopRandomly();
		else
			currentCity = selectNextStop(choice_info);
	} else {
		currentCity = selectNextStop(choice_info);
	}
	tourOrder[n_leg] = currentCity;
	notVisited[currentCity] = 0;
	n_leg ++;
}

vector<int> Ant::getTourOrder() {
	return tourOrder;
}

void Ant::setTourLength(double length) {
	tourLength = length;
}
/*
TourInfo Ant::makeTour(vector<double> &eta, double *choice_info) {
	for(int i = 0; i < n_city; i++ ) {
		notVisited[i] = 1;
	}
	tourLength = 0.0;

	bool randomize = params->isRandomize();
	double p_fancy = params->getP_fancy();
	
	if(randomize && (rand->nextDouble() < p_fancy))
		currentCity = selectStartPointRandomly();
	else
		currentCity = selectStartPoint(choice_info);
	tourOrder[0] = currentCity;
	notVisited[currentCity] = 0;
	
	for(int i = 1; i < n_city; i++ ) {
		int lastCity = currentCity;
		if(randomize && (rand->nextDouble() < p_fancy))
			currentCity = selectNextStopRandomly();
		else
			currentCity = selectNextStop(choice_info);
		tourOrder[i] = currentCity;
		notVisited[currentCity] = 0;
		tourLength += 1 / eta[lastCity * n_city + currentCity];
	}
	TourInfo tourInfo(tourLength, tourOrder);
	return tourInfo;
}
*/

void Ant::secretePheromone(double *tau) {
	double delta_tau = 0.0;
	if(tourLength > 0)
		delta_tau = params->getQ() / tourLength;
	for(int i = 1; i < n_city; i++ ) {
		tau[tourOrder[i - 1] * n_city + tourOrder[i]] += delta_tau;
		tau[tourOrder[i] * n_city + tourOrder[i - 1]] += delta_tau;
		//tau[tourOrder[i] * n_city + tourOrder[i - 1]] = tau[tourOrder[i - 1] * n_city + tourOrder[i]];
	}
	tau[tourOrder[0] * n_city + tourOrder[0]] += delta_tau;
	tau[tourOrder[n_city - 1] * n_city + tourOrder[n_city - 1]] += delta_tau;
}

int Ant::selectStartPoint(double *choice_info) {
	int startPoint = 0;
	double sum = 0.0;
	for(int i = 0; i < n_city; i++ ) {
		p[i] = choice_info[i * n_city + i];
		sum += p[i];
	}
	double value = sum * rand->nextDouble();
	sum = 0.0;
	for(int i = 0; i < n_city; i++ ) {
		sum += p[i];
		if(sum > value) {
			startPoint = i;
			break;
		}
	}
	return startPoint;
}

int Ant::selectStartPointRandomly() {
	int startPoint = 0;
	double sum = 0.0;
	for(int i = 0; i < n_city; i++ ) {
		p[i] = 1.0;
		sum += p[i];
	}
	double value = sum * rand->nextDouble();
	sum = 0.0;
	for(int i = 0; i < n_city; i++ ) {
		sum += p[i];
		if(sum > value) {
			startPoint = i;
			break;
		}
	}
	return startPoint;
}

int Ant::selectNextStop(double *choice_info) {
	int nextStop = 0;
	double sum = 0.0;
	for(int i = 0; i < n_city; i++ ) {
		p[i] = notVisited[i] * choice_info[currentCity * n_city + i];
		sum += p[i];
	}
	double value = sum * rand->nextDouble();
	sum = 0.0;
	for(int i = 0; i < n_city; i++ ) {
		sum += p[i];
		if(sum > value) {
			nextStop = i;
			break;
		}
	}
	return nextStop;
}

int Ant::selectNextStopRandomly() {
	int nextStop = 0;
	double sum = 0.0;
	for(int i = 0; i < n_city; i++ ) {
		p[i] = notVisited[i];
		sum += p[i];
	}
	double value = sum * rand->nextDouble();
	sum = 0.0;
	for(int i = 0; i < n_city; i++ ) {
		sum += p[i];
		if(sum > value) {
			nextStop = i;
			break;
		}
	}
	return nextStop;
}

int Ant::selectNextStopDetermin(double *choice_info) {
	double max = 0.0;
	int nextStop = 0;
	for(int i = 0; i < n_city; i++ )
		if(notVisited[i] * choice_info[currentCity * n_city + i] > max) {
			max = choice_info[currentCity * n_city + i];
			nextStop = i;
		}
	return nextStop;
}
