/*
AntMap: Software for constructing genetic linkage maps using an Ant Colony Optimization (ACO) algorithm
Copyright (C) 2004 National Agricultural and Bio-oriented Research Organization (NARO)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

Hiroyoshi IWATA
Department of Information Science and Technology
National Agricultural Research Center
3-1-1 Kannondai
Tsukuba
Ibaraki 305-8666
JAPAN

ph        +81-298-38-7025
fax       +81-298-38-8551
e-mail    iwatah@affrc.go.jp
*/

/////////////////////////////////////////////////////////////////////////////////////
#include "antcolony.h"
/////////////////////////////////////////////////////////////////////////////////////

//#include <iostream>

AntColony::AntColony(AcoParams *ap, double *d) {
	rand = new Random();
	init(ap, d);
};

AntColony::AntColony(AcoParams *ap, double *d, unsigned long seed) {
	rand = new Random(seed);
	init(ap, d);
};

AntColony::~AntColony() {
	if(rand) delete rand;
	if(tau) delete [] tau;
	if(choice_info) delete [] choice_info;
	for_each(ant.begin(), ant.end(), DeleteObject());
};

void AntColony::init(AcoParams *ap, double *d) {
	params = ap;
	dist = d;
	n_city = params->getN_city();
	tau = new double [n_city * n_city];
	choice_info = new double[n_city * n_city];
	eta_beta.resize(n_city * n_city);
	double init_tau = params->getInit_tau();
	for(int i = 0; i < n_city; i++ )
		for(int j = i; j < n_city; j++ ) {
			tau[i * n_city + j] = init_tau;
			tau[j * n_city + i] = init_tau;
		}
	double beta = params->getBeta();
	for(int i = 0; i < n_city; i++ ) {
		eta_beta[i * n_city + i] = 0.0;
		for(int j = i + 1; j < n_city; j++ ) {
			if(dist[i * n_city + j] > 0)
				eta_beta[i * n_city + j] = pow(1 / dist[i * n_city + j], beta);
			else
				eta_beta[i * n_city + j] = pow(1 / MINIMUM_DIST, beta);
			eta_beta[j * n_city + i] = eta_beta[i * n_city + j];
		}
	}
	n_ant = params->getN_ant();
	ant.reserve(n_ant);
	for(int i = 0; i < n_ant; i++ )
		ant.push_back(new Ant(params, rand));
}

TourInfo AntColony::findBestTour() {
	TourInfo tempBest;
	int n_replicate = params->getN_replicate();
	for(int i = 0; i < n_replicate; i++ ) {
		computeChoiceInfo();
		tempBest = sendAntsOnJourney();
		if(tempBest.getTourLength() < bestTour.getTourLength())
			bestTour = tempBest;
		updatePheromoneLevel();
	}
	return bestTour;
}

unsigned long AntColony::getNextSeed() {
	return rand->nextInt();
}

/*
TourInfo AntColony::sendAntsOnJourney() {
	TourInfo tempTour;
	TourInfo tempBest;

	for(int i = 0; i < n_ant; i++ ) {
		tempTour = ant[i]->makeTour(eta, choice_info);
		if(tempTour.getTourLength() < tempBest.getTourLength())
			tempBest = tempTour;
	}
	return tempBest;
}
*/

TourInfo AntColony::sendAntsOnJourney() {
	TourInfo tempBest;
	vector<int> tourOrder;
	double tourLength;
	
	for(int i = 0; i < n_ant; i++ ) {
		ant[i]->prepareForJourney();
		ant[i]->depart(choice_info);
	}

	for(int i = 1; i < n_city; i++ ) {
		for(int j = 0; j < n_ant; j++ ) {
			ant[j]->move(choice_info);
		}
	}
	
	for(int i = 0; i < n_ant; i++ ) {
		tourOrder = ant[i]->getTourOrder();
		tourLength = getTourLength(tourOrder);
		ant[i]->setTourLength(tourLength);
		if(tourLength < tempBest.getTourLength()) {
			tempBest.setTourOrder(tourOrder);
			tempBest.setTourLength(tourLength);
		}
	}
	
	return tempBest;
}

double AntColony::getTourLength(const vector<int> &tourOrder) {
	double tourLength = 0.0;
	for(int i = 1; i < n_city; i++ )
		tourLength += dist[tourOrder[i] * n_city + tourOrder[i - 1]];
	return tourLength;
}

void AntColony::updatePheromoneLevel() {
	double rho = params->getRho();
	for(int i = 0; i < n_city; i++ )
		/*
		for(int j = i; j < n_city; j++ ) {
			tau[i * n_city + j] *= rho;
			tau[j * n_city + i] *= rho;
			//tau[j * n_city + i] = tau[i * n_city + j];
		}
		*/
		for(int j = 0; j < n_city; j++ )
			tau[i * n_city + j] *= rho;
	for(int i = 0; i < n_ant; i++ )
		ant[i]->secretePheromone(tau);
	markBestTour();
}

void AntColony::computeChoiceInfo() {
	int alpha = params->getAlpha();
	for(int i = 0; i < n_city; i++ ) {
		choice_info[i * n_city + i] = pow(tau[i * n_city + i], alpha);
		for(int j = i + 1; j < n_city; j++ ) {
			choice_info[i * n_city + j] = pow(tau[i * n_city + j], alpha) * eta_beta[i * n_city + j];
			choice_info[j * n_city + i] = choice_info[i * n_city + j];
		}
	}
}

void AntColony::markBestTour() {
	double weight = params->getWeight();
	if(weight > 0) {
		double delta_tau = 0.0;
		double tourLength = bestTour.getTourLength();
		vector<int> tourOrder = bestTour.getTourOrder();
		if(tourLength > 0)
			delta_tau = params->getQ() / tourLength;
		for(int i = 1; i < n_city; i++ ) {
			tau[tourOrder[i - 1] * n_city + tourOrder[i]] += delta_tau * weight;
			tau[tourOrder[i] * n_city + tourOrder[i - 1]] += delta_tau * weight;
			//tau[tourOrder[i] * n_city + tourOrder[i - 1]] = tau[tourOrder[i - 1] * n_city + tourOrder[i]];

		}
		tau[tourOrder[0] * n_city + tourOrder[0]] += delta_tau * weight;
		tau[tourOrder[n_city - 1] * n_city + tourOrder[n_city - 1]] += delta_tau * weight;
	}
}
