/*
AntMap: Software for constructing genetic linkage maps using an Ant Colony Optimization (ACO) algorithm
Copyright (C) 2004 National Agricultural and Bio-oriented Research Organization (NARO)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

Hiroyoshi IWATA
Department of Information Science and Technology
National Agricultural Research Center
3-1-1 Kannondai
Tsukuba
Ibaraki 305-8666
JAPAN

ph        +81-298-38-7025
fax       +81-298-38-8551
e-mail    iwatah@affrc.go.jp
*/

package org.agmodel.bioinfo.antmap;

import org.agmodel.bioinfo.stat.rand.MersenneTwister;

/**
 * @author iwata
 *
 */
public class Ant {
	private AcoParamsBean params;
	//static private MersenneTwister rand = new MersenneTwister();
	//All ants use the same MerrsenneTwister object because it is defined as a static variable
	private MersenneTwister rand;
	private int [] notVisited;
	private double [] p;
	private int [] tourOrder;
	private double tourLength;
	private int n_city;
	private int currentCity;
	private int n_leg;

	/**
	 * Construct an Ant object
	 * @param params ACO parameters
	 * @param n_city Number of cities
	 */
	//public Ant(AcoParamsBean params, int n_city) {
	public Ant(AcoParamsBean params, int n_city, MersenneTwister rand) {
		this.params = params;  //ep[^
		this.rand = rand;
		this.n_city = n_city;  //ss̎擾
		notVisited = new int[n_city];  //KsstO̐ݒ
		for(int i = 0; i < n_city; i ++) {
			notVisited[i] = 1;
		}
		p = new double[n_city];  //ess̑I𗦁ia1ɊĂȂj
		tourOrder = new int [n_city];
		currentCity = 0;
	}

	/**
	 * Prepare for a journey <br>
	 * (Initialize flags for visited cities)
	 */
	public void prepareForJourney() {
		for(int i = 0; i < n_city; i++ )
			notVisited[i] = 1;
		n_leg = 0;
		tourLength = 0.0;
	}

	/**
	 * Set out on a journey
	 * @param choice_info
	 */
	public void depart(double [][] choice_info) {
		if(params.isRandomize()) {
			if(rand.nextDouble() < params.getP_fancy())
				currentCity = selectStartPointRandomly();
			else
				currentCity = selectStartPoint(choice_info);
		} else {
			currentCity = selectStartPoint(choice_info);
		}
		tourOrder[n_leg] = currentCity;
		notVisited[currentCity] = 0;
		n_leg ++;
	}

	/**
	 * Move to the next stop
	 * @param choice_info Heuristic information for choosing city
	 */
	public void move(double [][] choice_info) {
		if(params.isRandomize()) {
			if(rand.nextDouble() < params.getP_fancy())
				currentCity = selectNextStopRandomly();
			else
				currentCity = selectNextStop(choice_info);
		} else {
			currentCity = selectNextStop(choice_info);
		}
		tourOrder[n_leg] = currentCity;
		notVisited[currentCity] = 0;
		n_leg ++;
	}

	/**
	 * Get tour order
	 * @return Tour order
	 */
	public int [] getTourOrder() {
		//return tourOrder;
		int [] order = new int[tourOrder.length];
		for(int i = 0; i < order.length; i++ ) {
			order[i] = tourOrder[i];
		}
		return order;
	}

	/**
	 * Set tour length
	 * @param length Tour length
	 */
	public void setTourLength(double length) {
		tourLength = length;
	}

	/**
	 * Secrete pheromone on the tau matrix
	 */
	public void secretePheromone(double [][] tau) {
		double delta_tau = 0.0;
		if(tourLength > 0)
			delta_tau = params.getQ() / tourLength;
		for(int i = 1; i < n_city; i++ ) {
			tau[tourOrder[i - 1]][tourOrder[i]] += delta_tau;
			tau[tourOrder[i]][tourOrder[i - 1]] += delta_tau;
		}
		tau[tourOrder[0]][tourOrder[0]] += delta_tau; //mark starting point
		tau[tourOrder[n_city - 1]][tourOrder[n_city - 1]] += delta_tau; //mark end point
	}

	/**
	 * Choose a starting city
	 * @return Index number of city
	 */
	private int selectStartPoint(double [][] choice_info) {
		int startPoint = 0;
		double sum = 0.0;
		for(int i = 0; i < n_city; i++ ) {
			p[i] = choice_info[i][i];
			sum += p[i];
		}
		double value = sum * rand.nextDouble();
		sum = 0.0;
		for(int i = 0; i < n_city; i++ ) {
			sum += p[i];
			if(sum > value) {
				startPoint = i;
				break;
			}
		}
		return startPoint;
	}

	/**
	 * Choose a starting city randomly
	 * @return Index number of city
	 */
	private int selectStartPointRandomly() {
		int startPoint = 0;
		double sum = 0.0;
		for(int i = 0; i < n_city; i++ ) {
			p[i] = 1.0;
			sum += p[i];
		}
		double value = sum * rand.nextDouble();
		sum = 0.0;
		for(int i = 0; i < n_city; i++ ) {
			sum += p[i];
			if(sum > value) {
				startPoint = i;
				break;
			}
		}
		return startPoint;
	}

	/**
	 * Choose a next stop
	 * @return Index number of city
	 */
	private int selectNextStop(double [][] choice_info) {
		int nextStop = 0;
		double sum = 0.0;
		for(int i = 0; i < n_city; i++ ) {
			p[i] = notVisited[i] * choice_info[currentCity][i];
			sum += p[i];
		}
		double value = sum * rand.nextDouble();
		sum = 0.0;
		for(int i = 0; i < n_city; i++ ) {
			sum += p[i];
			if(sum > value) {
				nextStop = i;
				break;
			}
		}
		return nextStop;
	}

	/**
	 * Choose a next stop randomly
	 * @return Index number of city
	 */
	private int selectNextStopRandomly() {
		int nextStop = 0;
		double sum = 0.0;
		for(int i = 0; i < n_city; i++ ) {
			p[i] = notVisited[i];
			sum += p[i];
		}
		double value = sum * rand.nextDouble();
		sum = 0.0;
		for(int i = 0; i < n_city; i++ ) {
			sum += p[i];
			if(sum > value) {
				nextStop = i;
				break;
			}
		}
		return nextStop;
	}
}
