/*
AntMap: Software for constructing genetic linkage maps using an Ant Colony Optimization (ACO) algorithm
Copyright (C) 2004 National Agricultural and Bio-oriented Research Organization (NARO)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

Hiroyoshi IWATA
Department of Information Science and Technology
National Agricultural Research Center
3-1-1 Kannondai
Tsukuba
Ibaraki 305-8666
JAPAN

ph        +81-298-38-7025
fax       +81-298-38-8551
e-mail    iwatah@affrc.go.jp
*/

package org.agmodel.bioinfo.antmap;

import org.agmodel.bioinfo.stat.rand.MersenneTwister;

/**
 * @author iwata
 *
 */
public class AntColony implements AntColonyOptimizer {
	private AcoParamsBean params;
	private double [][] dist;
	private MersenneTwister rand;
	private double [][] tau;
	private double [][] choice_info;
	private double [][] eta_beta;
	private Ant [] ant;
	private int n_city;
	int n_ant;
	TourBean bestTour = new TourBean();
	private final double MINIMUM_DIST = 0.00000001;

	/**
	 * Construct an AntColony object (100% pure Java version of AntColony)
	 * @param params ACO parameters
	 * @param dist Distance matrix
	 */
	public AntColony(AcoParamsBean params, double [][] dist) {
		this(params, dist, System.currentTimeMillis());
	}


	/**
	 * Construct an AntColony object (100% pure Java version of AntColony)
	 * @param params ACO parameters
	 * @param dist Distance matrix
	 * @param seed Seed for a random number generator
	 */
	public AntColony(AcoParamsBean params, double [][] dist, long seed) {
		rand = new MersenneTwister(seed);
		this.params = params;
		this.dist = dist;
		n_city = dist.length;
		tau = new double [n_city][n_city];
		choice_info = new double [n_city][n_city];
		eta_beta = new double [n_city][n_city];
		double init_tau = params.getInit_tau();
		for(int i = 0; i < n_city; i++ ) {
			for(int j = i; j < n_city; j++ ) {
				tau[i][j] = init_tau;
				tau[j][i] = init_tau;
			}
		}
		double beta = params.getBeta();
		for(int i = 0; i < n_city; i++ ) {
			eta_beta[i][i] = 0.0;
			for(int j = i + 1; j < n_city; j++ ) {
				if(dist[i][j] > 0)
					eta_beta[i][j] = Math.pow(1 / dist[i][j], beta);
				else
					eta_beta[i][j] = Math.pow(1 / MINIMUM_DIST, beta);
				eta_beta[j][i] = eta_beta[i][j];
			}
		}
		n_ant = params.getN_ant();
		ant = new Ant[n_ant];
		for(int i = 0; i < n_ant; i++ ) {
			ant[i] = new Ant(params, n_city, rand);
		}
	}

	/**
	 * Find the best tour by using Ant Colony Optimization
	 * @ return Best tour
	 */
	public TourBean findBestTour() {
		TourBean tempBest;
		int n_iteration = params.getN_iteration();
		for(int i = 0; i < n_iteration; i++ ) {
			computeChoiceInfo();
			tempBest = sendAntsOnJourney();
			if(tempBest.getTourLength() < bestTour.getTourLength())
				bestTour = tempBest;
			updatePheromoneLevel();
		}
		return bestTour;
	}

	private TourBean sendAntsOnJourney() {
		TourBean tempBest = new TourBean();
		int [] tourOrder = new int [n_city];
		double tourLength;

		for(int i = 0; i < n_ant; i++ ) {
			ant[i].prepareForJourney();
			ant[i].depart(choice_info);
		}

		for(int i = 1; i < n_city; i++ ) {
			for(int j = 0; j < n_ant; j++ ) {
				ant[j].move(choice_info);
			}
		}

		for(int i = 0; i < n_ant; i++ ) {
			tourOrder = ant[i].getTourOrder();
			tourLength = getTourLength(tourOrder);
			ant[i].setTourLength(tourLength);
			if(tourLength < tempBest.getTourLength()) {
				tempBest.setTourOrder(tourOrder);
				tempBest.setTourLength(tourLength);
			}
		}

		return tempBest;
	}

	private double getTourLength(int [] tourOrder) {
		double tourLength = 0.0;
		for(int i = 1; i < n_city; i++ )
			tourLength += dist[tourOrder[i]][tourOrder[i - 1]];
		return tourLength;
	}

	private void updatePheromoneLevel() {
		double rho = params.getRho();
		for(int i = 0; i < n_city; i++ )
			for(int j = 0; j < n_city; j++ )
				tau[i][j] *= rho;
		for(int i = 0; i < n_ant; i++ )
			ant[i].secretePheromone(tau);
		markBestTour();
	}

	private void computeChoiceInfo() {
		double alpha = params.getAlpha();
		for(int i = 0; i < n_city; i++ ) {
			choice_info[i][i] = Math.pow(tau[i][i], alpha);
			for(int j = i + 1; j < n_city; j++ ) {
				choice_info[i][j] = Math.pow(tau[i][j], alpha) * eta_beta[i][j];
				choice_info[j][i] = choice_info[i][j];
			}
		}
	}

	private void markBestTour() {
		double weight = params.getWeight();
		if(weight > 0) {
			double delta_tau = 0.0;
			double tourLength = bestTour.getTourLength();
			int [] tourOrder = bestTour.getTourOrder();
			if(tourLength > 0)
				delta_tau = params.getQ() / tourLength;
			for(int i = 1; i < tourOrder.length; i++ ) {
				tau[tourOrder[i - 1]][tourOrder[i]] += delta_tau * weight;
				tau[tourOrder[i]][tourOrder[i - 1]] += delta_tau * weight;
			}
			tau[tourOrder[0]][tourOrder[0]] += delta_tau * weight; //mark starting point
			tau[tourOrder[n_city - 1]][tourOrder[n_city - 1]] += delta_tau * weight; //mark end point
		}
	}
}
