/*
AntMap: Software for constructing genetic linkage maps using an Ant Colony Optimization (ACO) algorithm
Copyright (C) 2004 National Agricultural and Bio-oriented Research Organization (NARO)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

Hiroyoshi IWATA
Department of Information Science and Technology
National Agricultural Research Center
3-1-1 Kannondai
Tsukuba
Ibaraki 305-8666
JAPAN

ph        +81-298-38-7025
fax       +81-298-38-8551
e-mail    iwatah@affrc.go.jp
*/

package org.agmodel.bioinfo.antmap;

import org.agmodel.bioinfo.segregation.*;

/**
 * @author iwata
 * gvZJavaBean
 * ̐ꂽRg̑}ev[gύX邽
 * EBhE > ݒ > Java > R[h > R[hƃRg
 */
public class LinkageAnalyzer {
	private TwoPointAnalysis tpa = new TwoPointAnalysis();
	private SegDataBean seg;
	private int [][] n = new int [1][1];
	private double [][] r = new double [1][1];
	private double [][] chi2 = new double [1][1];
	private double [][] p = new double [1][1];
	private double [][] lod = new double [1][1];
	private boolean execute = false;

	/**
	 * RXgN^
	 */
	public LinkageAnalyzer(SegDataBean seg) {
		setSegDataBean(seg);
	}

	/**
	 * SegDataBeañZb^[
	 * @param seg
	 */
	public void setSegDataBean(SegDataBean seg) {
		this.seg = seg;
		execute = false;
	}

	/**
	 * ǧvZ
	 * @return gs
	 */
	public void execute() {
		int segType = seg.getSegType();
		char [][] segData = seg.getGenotypeData();
		int n_marker = segData.length;
		n = new int [n_marker][n_marker];
		r = new double [n_marker][n_marker];
		chi2 = new double [n_marker][n_marker];
		p = new double [n_marker][n_marker];
		lod = new double [n_marker][n_marker];

		if(segType == SegType.ST_BC) {
			twoPoint_BC(segData);
		} else if(segType == SegType.ST_F2) {
			int [] markerType = seg.getMarkerType();
			twoPoint_F2(segData, markerType);
		} else if(segType == SegType.ST_DH) {
			twoPoint_DH(segData);
		} else if(segType == SegType.ST_RIL) {
			twoPoint_RIL(segData);
		}
		for(int i = 0; i < n_marker; i++ ) {
			for(int j = i + 1; j < n_marker; j++ ) {
				n[j][i] = n[i][j];
				r[j][i] = r[i][j];
				chi2[j][i] = chi2[i][j];
				p[j][i] = p[i][j];
				lod[j][i] = lod[i][j];
			}
		}
		execute = true;
	}

	/**
	 * ߂G
	 * @return gs
	 */
	private void twoPoint_BC(char [][] segData) {
		int n_marker = segData.length;
		RecombinationBean rb;

		for(int i = 0; i < n_marker; i++ ) {
			r[i][i] = 0.0;
			for(int j = i + 1; j < n_marker; j++ ) {
				rb = tpa.getRecombinationBean_BC(segData[i], segData[j]);
				n[i][j] = rb.getN();
				r[i][j] = rb.getValue();
				chi2[i][j] = rb.getChi2();
				p[i][j] = rb.getP();
				lod[i][j] = rb.getLod();
			}
		}
	}

	/**
	 * F2
	 * @return gs
	 */
	private void twoPoint_F2(char [][] segData, int [] markerType) {
		int n_marker = segData.length;
		RecombinationBean rb;

		for(int i = 0; i < n_marker; i++ ) {
			r[i][i] = 0.0;
			for(int j = i + 1; j < n_marker; j++ ) {
				rb = tpa.getRecombinationBean_F2(segData[i], segData[j], markerType[i], markerType[j]);
				n[i][j] = rb.getN();
				r[i][j] = rb.getValue();
				chi2[i][j] = rb.getChi2();
				p[i][j] = rb.getP();
				lod[i][j] = rb.getLod();
			}
		}
	}

	/**
	 * {
	 * @return gs
	 */
	private void twoPoint_DH(char [][] segData) {
		int n_marker = segData.length;
		RecombinationBean rb;

		for(int i = 0; i < n_marker; i++ ) {
			r[i][i] = 0.0;
			for(int j = i + 1; j < n_marker; j++ ) {
				rb = tpa.getRecombinationBean_DH(segData[i], segData[j]);
				n[i][j] = rb.getN();
				r[i][j] = rb.getValue();
				chi2[i][j] = rb.getChi2();
				p[i][j] = rb.getP();
				lod[i][j] = rb.getLod();
			}
		}
	}

	/**
	 * F2Ȓgߌni吢j
	 * @return gs
	 */
	private void twoPoint_RIL(char [][] segData) {
		int n_marker = segData.length;
		RecombinationBean rb;

		for(int i = 0; i < n_marker; i++ ) {
			r[i][i] = 0.0;
			for(int j = i + 1; j < n_marker; j++ ) {
				rb = tpa.getRecombinationBean_RIL(segData[i], segData[j]);
				n[i][j] = rb.getN();
				r[i][j] = rb.getValue();
				chi2[i][j] = rb.getChi2();
				p[i][j] = rb.getP();
				lod[i][j] = rb.getLod();
			}
		}
	}

	/**
	 * @return
	 */
	public int[][] getN() {
		if(!execute) {
			execute();
		}
		return n;
	}

	/**
	 * @return
	 */
	public double[][] getP() {
		if(!execute) {
			execute();
		}
		return p;
	}

	/**
	 * @return
	 */
	public double[][] getR() {
		if(!execute) {
			execute();
		}
		return r;
	}

	/**
	 * @return
	 */
	public SegDataBean getSeg() {
		if(!execute) {
			execute();
		}
		return seg;
	}
	/**
	 * @return
	 */
	public double[][] getLod() {
		if(!execute) {
			execute();
		}
		return lod;
	}

}
