/*
AntMap: Software for constructing genetic linkage maps using an Ant Colony Optimization (ACO) algorithm
Copyright (C) 2004 National Agricultural and Bio-oriented Research Organization (NARO)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

Hiroyoshi IWATA
Department of Information Science and Technology
National Agricultural Research Center
3-1-1 Kannondai
Tsukuba
Ibaraki 305-8666
JAPAN

ph        +81-298-38-7025
fax       +81-298-38-8551
e-mail    iwatah@affrc.go.jp
*/

package org.agmodel.bioinfo.antmap;

import org.agmodel.bioinfo.segregation.TwoPointAnalysis;
import org.agmodel.bioinfo.segregation.SegType;
import org.agmodel.bioinfo.stat.rand.MersenneTwister;
/**
 * @author iwata
 *
 */
public class RMatrixFactory {
	static private MersenneTwister rand = new MersenneTwister();
	TwoPointAnalysis tpa = new TwoPointAnalysis();

	public double [][] getR(SegDataBean segDataBean) {
		double [][] r = new double [1][1];
		int segType = segDataBean.getSegType();
		char [][] genotypeData = segDataBean.getGenotypeData();
		if(segType == SegType.ST_BC)
			r = convertSegData2R_BC(genotypeData);
		else if(segType == SegType.ST_DH)
			r = convertSegData2R_DH(genotypeData);
		else if(segType == SegType.ST_RIL)
			r = convertSegData2R_RIL(genotypeData);
		else if(segType == SegType.ST_F2)
			r = convertSegData2R_F2(genotypeData, segDataBean.getMarkerType());
		return r;
	}

	public double [][] getR(SegDataBean segDataBean, int [] markerIndex) {
		double [][] r = new double [1][1];
		int segType = segDataBean.getSegType();
		char [][] genotypeData = new PartialExtractor().extractPartialGenotypeData(segDataBean.getGenotypeData(), markerIndex);
		if(segType == SegType.ST_BC)
			r = convertSegData2R_BC(genotypeData);
		else if(segType == SegType.ST_DH)
			r = convertSegData2R_DH(genotypeData);
		else if(segType == SegType.ST_RIL)
			r = convertSegData2R_RIL(genotypeData);
		else if(segType == SegType.ST_F2) {
			int [] markerType = new PartialExtractor().extractParitalArray(segDataBean.getMarkerType(), markerIndex);
			r = convertSegData2R_F2(genotypeData, markerType);
		}
		return r;
	}

	public double [][] getBootStrappedR(SegDataBean segDataBean) {
		double [][] r = new double [1][1];
		int segType = segDataBean.getSegType();
		char [][] genotypeData = bootStrap(segDataBean.getGenotypeData());
		if(segType == SegType.ST_BC)
			r = convertSegData2R_BC(genotypeData);
		else if(segType == SegType.ST_DH)
			r = convertSegData2R_DH(genotypeData);
		else if(segType == SegType.ST_RIL)
			r = convertSegData2R_RIL(genotypeData);
		else if(segType == SegType.ST_F2)
			r = convertSegData2R_F2(genotypeData, segDataBean.getMarkerType());
		return r;
	}

	public double [][] getBootStrappedR(SegDataBean segDataBean, int [] markerIndex) {
		double [][] r = new double [1][1];
		int segType = segDataBean.getSegType();
		char [][] genotypeData = bootStrap(
			new PartialExtractor().extractPartialGenotypeData(
					segDataBean.getGenotypeData(), markerIndex)
			);
		if(segType == SegType.ST_BC)
			r = convertSegData2R_BC(genotypeData);
		else if(segType == SegType.ST_DH)
			r = convertSegData2R_DH(genotypeData);
		else if(segType == SegType.ST_RIL)
			r = convertSegData2R_RIL(genotypeData);
		else if(segType == SegType.ST_F2) {
			int [] markerType = new PartialExtractor().extractParitalArray(segDataBean.getMarkerType(), markerIndex);
			r = convertSegData2R_F2(genotypeData, markerType);
		}
		return r;
	}

	public char [][] bootStrap(char [][] genotypeData) {
		char [][] bootData = new char [genotypeData.length][genotypeData[0].length];
		int [] randomSelector = new int[genotypeData[0].length];

		for(int i = 0; i < randomSelector.length; i++ )
			randomSelector[i] = rand.nextInt(genotypeData[0].length);

		for(int i = 0; i < genotypeData.length; i++ )
			for(int j = 0; j < genotypeData[0].length; j++ )
				bootData[i][j] = genotypeData[i][randomSelector[j]];

		return bootData;
	}

	public double [][] convertSegData2R_BC(char [][] genotypeData) {
		double [][] r = new double [genotypeData.length][genotypeData.length];
		for(int i = 0; i < genotypeData.length; i++ )
			for(int j = 0; j < genotypeData.length; j++ )
				r[i][j] = tpa.getR_BC(genotypeData[i], genotypeData[j]);
		return r;
	}

	public double [][] convertSegData2R_DH(char [][] genotypeData) {
		double [][] r = new double [genotypeData.length][genotypeData.length];
		for(int i = 0; i < genotypeData.length; i++ )
			for(int j = 0; j < genotypeData.length; j++ )
				r[i][j] = tpa.getR_DH(genotypeData[i], genotypeData[j]);
		return r;
	}

	public double [][] convertSegData2R_RIL(char [][] genotypeData) {
		double [][] r = new double [genotypeData.length][genotypeData.length];
		for(int i = 0; i < genotypeData.length; i++ )
			for(int j = 0; j < genotypeData.length; j++ )
				r[i][j] = tpa.getR_RIL(genotypeData[i], genotypeData[j]);
		return r;
	}

	public double [][] convertSegData2R_F2(char [][] genotypeData, int [] markerType) {
		double [][] r = new double [genotypeData.length][genotypeData.length];
		for(int i = 0; i < genotypeData.length; i++ )
			for(int j = 0; j < genotypeData.length; j++ )
				r[i][j] = tpa.getR_F2(genotypeData[i], genotypeData[j], markerType[i], markerType[j]);
		return r;
	}
}
