/*
AntMap: Software for constructing genetic linkage maps using an Ant Colony Optimization (ACO) algorithm
Copyright (C) 2004 National Agricultural and Bio-oriented Research Organization (NARO)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

Hiroyoshi IWATA
Department of Information Science and Technology
National Agricultural Research Center
3-1-1 Kannondai
Tsukuba
Ibaraki 305-8666
JAPAN

ph        +81-298-38-7025
fax       +81-298-38-8551
e-mail    iwatah@affrc.go.jp
*/

package org.agmodel.bioinfo.antmap;

import java.io.FileReader;
import java.io.BufferedReader;
import java.util.StringTokenizer;
import java.util.Map;
import java.util.HashMap;
import org.agmodel.bioinfo.common.FileFormatException;
import org.agmodel.bioinfo.segregation.SegType;
import org.agmodel.bioinfo.segregation.MarkerTypeGetter;

/**
 * @author iwata
 *
 */
public class SegDataReader {
	private String fileName;
	private FileReader fr;

	public SegDataReader(String fileName) throws Exception {
		fr = new FileReader(fileName);
		this.fileName = fileName;
	}

	/**
	 *
	 * @param fileType
	 * @return
	 * @throws Exception
	 */
	public SegDataBean getSegData(String fileType) throws Exception {
		SegDataBean segDataBean = new SegDataBean();
		if(fileType.equalsIgnoreCase("MAPMAKER")) {
			segDataBean = readMapmakerData();
		}
		return segDataBean;
	}

	private SegDataBean readMapmakerData() throws Exception {
		int segType = SegType.ST_NA;
		int n_marker = 0;
		int n_individual = 0;
		int n_qtl = 0;
		char [][] genotypeData = new char[1][1];
		String [] markerName = new String [1];
		boolean caseFlag = false;
		boolean symbolsFlag = false;
		boolean readFirstLine = false;
		boolean readSecondLine = false;
		Map convertMap = new HashMap();

		try {
			BufferedReader br = new BufferedReader(fr);

			int lineno = 0;
			int markno = 0;
			int count = 0;
			while(br.ready()) {
				String line = br.readLine();
				lineno ++;

				if(line.length() == 0) continue; //s
				if(line.startsWith("#")) continue; //Rg

				StringTokenizer st = new StringTokenizer(line);

				if(!readFirstLine) {
					if(!st.nextToken().equalsIgnoreCase("data") || !st.nextToken().equalsIgnoreCase("type")) {
						throw new FileFormatException("Error in line" + lineno + " (cannot find 'data type' statement)");
					}
					String temp = st.nextToken();
					if(temp.equalsIgnoreCase("f2")) {
						if(!st.hasMoreTokens()) {
							throw new FileFormatException("Error in line" + lineno + " (unknown data type)");
						}
						temp = st.nextToken();
						if(temp.equalsIgnoreCase("intercross")) {
							segType = SegType.ST_F2;
						} else if(temp.equalsIgnoreCase("backcross")) {
							segType = SegType.ST_BC;
						} else {
							throw new FileFormatException("Error in line" + lineno + " (unknown data type)");
						}
					} else if(temp.equalsIgnoreCase("ri")) {
						if(!st.hasMoreTokens()) {
							throw new FileFormatException("Error in line" + lineno + " (unknown data type)");
						}
						temp = st.nextToken();
						if(temp.equalsIgnoreCase("self")) {
							segType = SegType.ST_RIL;
						} else if(temp.equalsIgnoreCase("sib")) {
							throw new FileFormatException("Sorry. Segregation type 'ri sib' is not supported!");
						} else {
							throw new FileFormatException("Error in line" + lineno + " (unknown data type)");
						}
					} else if(temp.equalsIgnoreCase("f3")) {
						if(!st.hasMoreTokens()) {
							throw new FileFormatException("Error in line" + lineno + " (unknown data type)");
						}
						temp = st.nextToken();
						if(temp.equalsIgnoreCase("self")) {
							throw new FileFormatException("Sorry. Segregation type 'f3 self' is not supported!");
						} else {
							throw new FileFormatException("Error in line" + lineno + " (unknown data type)");
						}
					} else if(temp.equalsIgnoreCase("dh")) {
						segType = SegType.ST_DH;
					} else {
						throw new FileFormatException("Error in line" + lineno + " (unknown data type)");
					}
					readFirstLine = true;
					continue;
				}

				if(!readSecondLine) {
					n_individual = new Integer(st.nextToken()).intValue();
					n_marker = new Integer(st.nextToken()).intValue();
					n_qtl = new Integer(st.nextToken()).intValue();

					while(st.hasMoreTokens()) {
						String temp = st.nextToken();
						if(temp.equalsIgnoreCase("case")) {
							caseFlag = true;
						} else if(temp.equalsIgnoreCase("symbols")) {
							symbolsFlag = true;
						} else if(symbolsFlag && temp.indexOf('=') == 1 && temp.length() == 3)	{
							convertMap.put(new Character(temp.charAt(0)), new Character(temp.toUpperCase().charAt(2)));
						} else {
							throw new FileFormatException("Error in line" + lineno + " (wrong header lines)");
						}
					}

					genotypeData = new char [n_marker][n_individual];
					markerName = new String [n_marker];
					readSecondLine = true;
					continue;
				}

				if(readFirstLine && readSecondLine && markno < n_marker) {
					if(line.startsWith("*")) {
						if(count > 0) { //f[^ȂƂ
							throw new FileFormatException("Error in line" + (lineno - 1) + " (the number of data is short)");
						}
						markerName[markno] = st.nextToken().substring(1);
					}
					while(st.hasMoreTokens()) {
						String temp = st.nextToken();
						for(int i = 0; i < temp.length(); i++ ) {
							if(count == n_individual) {
								throw new FileFormatException("Error in line" + lineno + " (the number of data is too much)");
							}
							char c = temp.charAt(i);
							if(!caseFlag) {
								c = Character.toUpperCase(c);
							}
							if(symbolsFlag && convertMap.containsKey(new Character(c))) {
								genotypeData[markno][count] = ((Character)convertMap.get(new Character(c))).charValue();
							} else {
								genotypeData[markno][count] = c;
							}
							count ++;
						}
					}
					if(count == n_individual) {
						markno ++;
						count = 0;
					}
				}
			}

			fr.close();

		} catch (Exception e) {
			throw e;
		}

		SegDataBean segDataBean = new SegDataBean(fileName, segType, markerName, genotypeData);
		if(segType == SegType.ST_F2)
				segDataBean.setMarkerType(new MarkerTypeGetter().getMarkerTypeArray(genotypeData));
		return segDataBean;
	}

}
