/*
 AntMap: Software for constructing genetic linkage maps using an Ant Colony Optimization (ACO) algorithm
 Copyright (C) 2004 National Agricultural and Bio-oriented Research Organization (NARO)

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

 Hiroyoshi IWATA
 Department of Information Science and Technology
 National Agricultural Research Center
 3-1-1 Kannondai
 Tsukuba
 Ibaraki 305-8666
 JAPAN

 ph        +81-298-38-7025
 fax       +81-298-38-8551
 e-mail    iwatah@affrc.go.jp
 */

package org.agmodel.bioinfo.antmap.gui;

import java.text.*;
import java.util.*;
import javax.swing.*;
import org.agmodel.bioinfo.antmap.*;
//import java.io.FileReader;
//import java.io.BufferedReader;
import org.agmodel.bioinfo.common.*;
import org.agmodel.bioinfo.segregation.*;

/**
 * A main engine for calculation in AntMap
 * <p>Title: AntMap</p>
 * <p>Explanation: Linkage mapping software based on ACO algorithm</p>
 * <p>Copyright: Copyright (c) 2004 NARO</p>
 * <p>Company name: NARO</p>
 * @author Hiroyoshi IWATA
 * @version 1.1
 */
public class AntMapEngine {
  SegDataBean segDataBean;
  GroupExtractor groupExtractor;
  MapData[] mapData;
  JTextArea log;
  JTextArea result;
  DrawMapPanel map;
  JProgressBar progress;
  JLabel status;
  ThreadIndicator threadIndicator = new ThreadIndicator();

  public AntMapEngine(MainFrame mainFrame) {
    this.log = mainFrame.myTextArea_log;
    this.result = mainFrame.myTextArea_result;
    this.map = mainFrame.drawMapPanel;
    this.progress = mainFrame.jProgressBar;
    this.status = mainFrame.jLabel_Status;
    threadIndicator.setStatusLabel(mainFrame.jLabel_Status);
  }

  public void loadData(String fileName) throws Exception {
    if (threadIndicator.isInService()) {
      throw new AntMapEngineException("Please be patiant! Under calculation.");
    }

    //initialize old data
    segDataBean = null;
    groupExtractor = null;
    mapData = null;

    //open and load data file
    log.append("--------open and verify data-------------------------------\n");
    log.append("open \"" + fileName + "\" ... ");
    try {
      SegDataReader segDataReader = new SegDataReader(fileName);
      segDataBean = segDataReader.getSegData("Mapmaker");
    }
    catch (Exception ex) {
      log.append("\n    Error: " + ex + "\n");
      throw ex;
    }
    log.append("ok\n");

    //show summary of data
    log.append("    Generation: " +
               SegType.getSegTypeName(segDataBean.getSegType()) + "\n");
    log.append("    # progenies: " + segDataBean.getN_individual() + "\n");
    log.append("    # markers: " + segDataBean.getN_marker() + "\n");

    //verify data
    log.append("verify data ... ");
    String[] error = new SegDataVerifier().verify(segDataBean);
    if (error.length > 0) {
      log.append("\n    " + error.length +
                 " error(s) found in segregation data\n");
      for (int i = 0; i < error.length; i++) {
        log.append("    " + error[i] + "\n");
      }
      segDataBean = null;
      throw new AntMapEngineException("Segregation data in " + fileName +
                                      " is in wrong format");
    }
    else {
      log.append("ok\n");

      //output to result panel
    }
    Date date = new Date();
    DateFormat dateFormat = DateFormat.getDateTimeInstance(DateFormat.FULL,
        DateFormat.LONG, Locale.US);

    result.append("--------------------------------------------------\n");
    result.append("Date and Time: " + dateFormat.format(date) + "\n");
    result.append("File Name: " + segDataBean.getFileName() + "\n");
    result.append("Generation: " +
                  SegType.getSegTypeName(segDataBean.getSegType()) + "\n");
    result.append("No of progenies: " + segDataBean.getN_individual() + "\n");
    result.append("No of markers: " + segDataBean.getN_marker() + "\n");
  }

  public void testSegregation() throws Exception {
    if (threadIndicator.isInService()) {
      throw new AntMapEngineException("Please be patiant! Under calculation.");
    }
    if (segDataBean == null) {
      throw new AntMapEngineException("Valid data have not been loaded yet");
    }

    log.append("--------segregation test-----------------------------------\n");
    log.append("test segregation ... ");
    SegRatioTester segRatioTester = new SegRatioTester();
    String[] message = segRatioTester.testAllLoci(segDataBean);
    log.append("ok\n");
    log.append("    " + message.length + " loci are tested\n");

    result.append("---------------------\n");
    result.append("Segregation Test:\n");
    result.append("Marker       N       Ratio    Chi^2   P\n");
    for (int i = 0; i < message.length; i++) {
      result.append(message[i] + "\n");
    }
  }

  public int linkageGrouping(GroupingOptionBean groupingOptionBean) throws
      Exception {
    if (threadIndicator.isInService()) {
      throw new AntMapEngineException("Please be patiant! Under calculation.");
    }
    if (segDataBean == null) {
      throw new AntMapEngineException("Valid data have not been loaded yet");
    }

    log.append("--------linkage grouping-----------------------------------\n");
    log.append("linkage grouping ... ");
    int method = groupingOptionBean.getMethod();
    int criterion = groupingOptionBean.getCriterion();
    double threshold = groupingOptionBean.getThreshold();
    int mapFunction = groupingOptionBean.getMapFunction();
    int minNMarker = groupingOptionBean.getMinNMarker();
    LGFactory lgFactory = new LGFactory();
    lgFactory.setMapFunction(mapFunction);
    groupExtractor = lgFactory.getGroupExtractor(segDataBean,
                                                 method,
                                                 criterion,
                                                 threshold,
                                                 minNMarker);
    lgFactory = null;
    int n_group = groupExtractor.getN_group();
    log.append("ok\n");
    log.append("    # of groups: " + n_group + "\n");

    result.append("---------------------\n");
    result.append("Linkage grouping:\n");
    result.append("Grouping method: " + new LGFactory().getMethodName(method) +
                  "\n");
    result.append("Grouping criterion: " +
                  new LGFactory().getCriterionName(criterion) + "\n");
    if (groupingOptionBean.getCriterion() == LGFactory.CR_DIST) {
      result.append("Map function: " + MapFunction.getFunctionName(mapFunction) +
                    "\n");
    }
    result.append("Grouping threshold: " + threshold + "\n");
    result.append("Minimum number of markers for each group: " + minNMarker +
                  "\n\n");
    result.append("No of groups: " + n_group + "\n");

    StringFormat stringFormat = new StringFormat();
    result.append("Groups    # markers\n");
    for (int i = 0; i < n_group; i++) {
      result.append(stringFormat.format("Group" + (i + 1) + ":", 10, true)
                    +
                    stringFormat.format(new Integer(groupExtractor.
          getN_marker(i)).toString(), 10, false) + "\n");
    }
    for (int i = 0; i < n_group; i++) {
      int[] markerIndex = groupExtractor.getMarkerIndices(i);
      result.append(stringFormat.format("Group" + (i + 1) + ":", 10, true)
                    +
                    stringFormat.format( (new Integer(markerIndex.length).toString() +
                                          " markers"), 12, true) + "\n");
      for (int j = 0; j < markerIndex.length; j++) {
        result.append(" " + markerIndex[j]);
      }
      result.append("\n");
    }
    return n_group;
  }

  public void locusOrdering(OrderingOptionBean orderingOptionBean,
                            AcoParamsBean acoParamsBean) throws Exception {
    if (threadIndicator.isInService()) {
      throw new AntMapEngineException("Please be patiant! Under calculation.");
    }
    if (groupExtractor == null) {
      throw new AntMapEngineException("Valid groups have not been identified");
    }

    log.append("--------locus ordering-------------------------------------\n");
    log.append("locus ordering ...\n");

    result.append("---------------------\n");
    result.append("Locus ordering:\n");
    result.append("Parameters for locus ordering\n");
    result.append("  Map function: " +
                  MapFunction.getFunctionName(orderingOptionBean.getMapFunction()) +
                  "\n");
    result.append("  Ordering criterion: " +
                  OrderingCriterion.getCriterionName(orderingOptionBean.
        getCriterion()) + "\n");
    result.append("  Max number of iterations: " +
                  orderingOptionBean.getMaxIteration() + "\n");
    result.append("  Number of iterations to confirm order: " +
                  orderingOptionBean.getNConfirm() + "\n");
    result.append("  Targeted group(s): ");
    int groupIndex = orderingOptionBean.getGroupIndex();
    if (groupIndex == -1) {
      result.append("All\n");
    }
    else {
      result.append("Group" + (groupIndex + 1) + "\n");
    }
    result.append("Parameters for ant colony optimization\n");
    result.append("  Alpha: " + acoParamsBean.getAlpha() + "\n");
    result.append("  Beta: " + acoParamsBean.getBeta() + "\n");
    result.append("  Rho: " + acoParamsBean.getRho() + "\n");
    result.append("  Q: " + acoParamsBean.getQ() + "\n");
    result.append("  No of iterations: " + acoParamsBean.getN_iteration() +
                  "\n");
    result.append("  No of ants: " + acoParamsBean.getN_ant() + "\n");
    result.append("  Initial value of tau: " + acoParamsBean.getInit_tau() +
                  "\n");
    result.append("  Probability of fanciful choice: " +
                  acoParamsBean.getP_fancy() + "\n");
    result.append("  Weight for the best tour: " + acoParamsBean.getWeight() +
                  "\n");

    int n_group = groupExtractor.getN_group();
    if (mapData == null) {
      mapData = new MapData[n_group];
    }
    Thread thread = new AntColonyOrdering(log,
                                          result,
                                          map,
                                          progress,
                                          mapData,
                                          segDataBean,
                                          groupExtractor,
                                          orderingOptionBean,
                                          acoParamsBean,
                                          threadIndicator);
    thread.start();
  }

  /*
     public void showMapData() {
    MapFunction mf = new MapFunction();
    System.out.println("Dist = " + mf.getHaldaneDistance(0.1));
    if(mapData != null) {
      for(int i = 0; i < mapData.length; i++ ) {
        if(mapData[i] != null)
          mapData[i].show();
        else
          System.out.println("Group " + (i + 1) + " is empty");
      }
    }
     }
   */

  public void bootstrapTest(
      OrderingOptionBean orderingOptionBean,
      AcoParamsBean acoParamsBean) throws Exception {
    if (threadIndicator.isInService()) {
      throw new AntMapEngineException("Please be patiant! Under calculation.");
    }
    if (mapData == null) {
      throw new AntMapEngineException("Valid groups have not been identified");
    }

    log.append("--------bootstrap test-------------------------------------\n");
    log.append("Bootstrap test ...\n");

    result.append("---------------------\n");
    result.append("Bootstrap test:\n");
    result.append("Parameters for locus ordering\n");
    result.append("  Map function: " +
                  MapFunction.getFunctionName(orderingOptionBean.getMapFunction()) +
                  "\n");
    result.append("  Ordering Criterion: " +
                  OrderingCriterion.getCriterionName(orderingOptionBean.
        getCriterion()) + "\n");
    result.append("  Max number of iterations: " +
                  orderingOptionBean.getMaxIteration() + "\n");
    result.append("  Number of iterations to confirm order: " +
                  orderingOptionBean.getNConfirm() + "\n");
    result.append("  # boostrap iterations: " +
                  orderingOptionBean.getNBootstrap() + "\n");
    result.append("  Targeted group(s): ");
    int groupIndex = orderingOptionBean.getGroupIndex();
    if (groupIndex == -1) {
      result.append("All\n");
    }
    else {
      result.append("Group" + (groupIndex + 1) + "\n");
    }
    result.append("Parameters for ant colony optimization\n");
    result.append("  Alpha: " + acoParamsBean.getAlpha() + "\n");
    result.append("  Beta: " + acoParamsBean.getBeta() + "\n");
    result.append("  Rho: " + acoParamsBean.getRho() + "\n");
    result.append("  Q: " + acoParamsBean.getQ() + "\n");
    result.append("  No of iterations: " + acoParamsBean.getN_iteration() +
                  "\n");
    result.append("  No of ants: " + acoParamsBean.getN_ant() + "\n");
    result.append("  Initial value of tau: " + acoParamsBean.getInit_tau() +
                  "\n");
    result.append("  Probability of fanciful choice: " +
                  acoParamsBean.getP_fancy() + "\n");
    result.append("  Weight for the best tour: " + acoParamsBean.getWeight() +
                  "\n");

    map.getDrawMapOptionBean().setShowBootstrapValue(true);
    Thread thread = new BootstrapTest(log,
                                      result,
                                      map,
                                      progress,
                                      mapData,
                                      segDataBean,
                                      orderingOptionBean,
                                      acoParamsBean,
                                      threadIndicator);
    thread.start();
  }

  public void redrawMap() throws Exception {
    if (threadIndicator.isInService()) {
      throw new AntMapEngineException("Please be patiant! Under calculation.");
    }
    if (mapData == null) {
      throw new AntMapEngineException("Valid groups have not been identified");
    }

    log.append("--------redraw map-----------------------------------------\n");
    log.append("Redraw map ... ");
    map.drawMap(mapData);
    log.append("ok\n");
  }

}
