/*
AntMap: Software for constructing genetic linkage maps using an Ant Colony Optimization (ACO) algorithm
Copyright (C) 2004 National Agricultural and Bio-oriented Research Organization (NARO)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

Hiroyoshi IWATA
Department of Information Science and Technology
National Agricultural Research Center
3-1-1 Kannondai
Tsukuba
Ibaraki 305-8666
JAPAN

ph        +81-298-38-7025
fax       +81-298-38-8551
e-mail    iwatah@affrc.go.jp
*/

package org.agmodel.bioinfo.segregation;

import java.util.List;
import java.util.LinkedList;
import java.util.Iterator;
import java.util.Random;

/**
 * @author iwata
 *
 */
public class Chromosome {
	private String name;
	private double length;
	private int nMarker;
	private List markerArray;
	private Random random;

	public Chromosome() {
		this("NA");
	}

	public Chromosome(String name) {
		this.name = name;
		nMarker = 0;
		markerArray = new LinkedList();
		random = new Random();
	}

	public Marker [] getMarkerArray() {
		Marker [] marker = new Marker [markerArray.size()];
		Iterator iterator = markerArray.iterator();
		int counter = 0;
		while(iterator.hasNext()) {
			marker[counter] = (Marker) iterator.next();
		}
		return marker;
	}

	public String [] getMarkerNameArray() {
		String [] markerName = new String [markerArray.size()];
		Iterator iterator = markerArray.iterator();
		int counter = 0;
		while(iterator.hasNext()) {
			markerName[counter] = ((Marker) iterator.next()).getName();
			counter ++;
		}
		return markerName;
	}

	public int [] getMarkerTypeArray() {
		int [] markerType = new int [markerArray.size()];
		Iterator iterator = markerArray.iterator();
		int counter = 0;
		while(iterator.hasNext()) {
			markerType[counter] = ((Marker) iterator.next()).getType();
			counter ++;
		}
		return markerType;
	}

	public double [] getMarkerLocationArray() {
		double [] markerLocation = new double [markerArray.size()];
		Iterator iterator = markerArray.iterator();
		int counter = 0;
		while(iterator.hasNext()) {
			markerLocation[counter] = ((Marker) iterator.next()).getLocation();
		}
		return markerLocation;
	}

	public int [] getPhenotype(int segType, double missingRate) {
		int [] phenotype = getPhenotype(segType);
		if(missingRate > 0.0) {
			for(int i = 0; i < phenotype.length; i++ ) {
				if(random.nextDouble() <= missingRate)
					phenotype[i] = 0;
			}
		}
		return phenotype;
	}

	public int [] getPhenotype(int segType) {
		int [] zygote = getZygote(segType);
		int [] markerType = getMarkerTypeArray();
		int length = Math.min(zygote.length, markerType.length);
		int [] phenotype = new int [length];
		for(int i = 0; i < length; i++ ) {
			if(segType == SegType.ST_F2) {
				if(markerType[i] == MarkerType.MT_DOM_P1) {
					if(zygote[i] != 2) {
						phenotype[i] = 5;
					} else {
						phenotype[i] = 2;
					}
			 	} else if(markerType[i] == MarkerType.MT_DOM_P2) {
					if(zygote[i] != 1) {
						phenotype[i] = 4;
					} else {
						phenotype[i] = 1; //2
					}
				} else {
					phenotype[i] = zygote[i];
				}
			} else { //BC, DH etc
				phenotype[i] = zygote[i];
			}
		}
		return phenotype;
	}

	public int [] getZygote(int segType) {
		int [] zygote = new int [markerArray.size()];
		if(segType == SegType.ST_F2) {
			int [] gamete1 = getGamete();
			int [] gamete2 = getGamete();
			for(int i = 0; i < gamete1.length; i++ ) {
				if(gamete1[i] != gamete2[i])
					zygote[i] = 3;
				else
					zygote[i] = gamete1[i];
			}
		} else if(segType == SegType.ST_BC) {
			int [] gamete1 = getGamete();
			zygote = new int [gamete1.length];
			for(int i = 0; i < gamete1.length; i++ ) {
				if(gamete1[i] == 2)
					zygote[i] = 3;
				else
					zygote[i] = gamete1[i];

			}
		} else if(segType == SegType.ST_DH) {
			zygote = getGamete();
		}
		return zygote;

	}

	public int [] getGamete() {
		Iterator iterator = markerArray.iterator();
		boolean [] boolGamete = new boolean [markerArray.size()];
		boolean prevState = random.nextBoolean();
		double prevLocation = 0.0;
		int counter = 0;
		while(iterator.hasNext()) {
			Marker marker = (Marker) iterator.next();
			double distance = marker.getLocation() - prevLocation;
			if(random.nextDouble() < distance)
				boolGamete[counter] = !prevState;
			else
				boolGamete[counter] = prevState;
			prevState = boolGamete[counter];
			prevLocation = marker.getLocation();
			counter ++;
		}
		int [] gamete = new int [boolGamete.length];
		for(int i = 0; i < boolGamete.length; i++ ) {
			if(boolGamete[i] == true)
				gamete[i] = 1;
			else
				gamete[i] = 2;
		}
		return gamete;
	}

	public void addMarker(Marker marker, double location) {
		Iterator iterator = markerArray.iterator();
		int counter = 0;
		while(iterator.hasNext()) {
			double l = ((Marker)iterator.next()).getLocation();
			if(l > location)
				break;
			counter ++;
		}
		marker.setLocation(location);
		markerArray.add(counter, marker);
		if(location > length)
			length = location;
		nMarker ++;
	}

	public void show() {
		Iterator iterator = markerArray.iterator();
		int counter = 0;
		while(iterator.hasNext()) {
			counter ++;
			Marker marker = (Marker)iterator.next();
			System.out.println(counter + "\t" + marker.getName() + "\t" + marker.getType() + "\t" + marker.getLocation());
		}
		System.out.println("Total length: " + length);
		System.out.println("Total # markers: " + nMarker);
	}

	public void showPhenotype(int segType, double missingRate) {
		int [] phenotype = getPhenotype(segType, missingRate);
		for(int i = 0; i < phenotype.length; i++ ) {
			System.out.print(phenotype[i]);
		}
		System.out.println();
	}

	public void showPhenotype(int segType) {
		showPhenotype(segType, 0.0);
	}

	public void showZygote(int segType) {
		int [] zygote = getZygote(segType);
		for(int i = 0; i < zygote.length; i++ ) {
			System.out.print(zygote[i]);
		}
		System.out.println();
	}

	public void showGamete() {
		int [] gamete = getGamete();
		for(int i = 0; i < gamete.length; i++ ) {
			System.out.print(gamete[i]);
		}
		System.out.println();
	}

	/**
	 * @return
	 */
	public double getLength() {
		return length;
	}

	/**
	 * @return
	 */
	public String getName() {
		return name;
	}

	/**
	 * @return
	 */
	public int getNMarker() {
		return nMarker;
	}

}
