/*
AntMap: Software for constructing genetic linkage maps using an Ant Colony Optimization (ACO) algorithm
Copyright (C) 2004 National Agricultural and Bio-oriented Research Organization (NARO)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

Hiroyoshi IWATA
Department of Information Science and Technology
National Agricultural Research Center
3-1-1 Kannondai
Tsukuba
Ibaraki 305-8666
JAPAN

ph        +81-298-38-7025
fax       +81-298-38-8551
e-mail    iwatah@affrc.go.jp
*/

package org.agmodel.bioinfo.segregation;

/**
 * @author iwata
 *
 */
public class FisherInfoForRBean implements java.io.Serializable {

	/**
	 * RXgN^
	 *
	 */
	public FisherInfoForRBean() {
	}

	/**
	 * Calculate Fisher's amount of information between codominant markers in F2 generation
	 * @param n # of individuals
	 * @param r Recombination rate
	 * @return Fisher's amount of information
	 */
	public double getFisherInfo_F2_CC(double n, double r) {
		if(r == 0) return Double.MAX_VALUE;
		return n * 2 * (1 - 3 * r + 3 * r * r) / (r * (1 - r) * (1 - 2 * r + 2 * r * r));
	}

	/**
	 * Calculate standard deviation of recombination rate between codomiannt markers in F2 generation
	 * @param n # of individuals
	 * @param r Recombination rate
	 * @return ST error of recombinatino rate
	 */
	public double getStd_F2_CC(double n, double r) {
		return 1 / Math.sqrt(getFisherInfo_F2_CC(n, r));
	}

	/**
	 * Calculate Fisher's amount of information between codominant and domiant markers in F2 generation
	 * @param n # of individuals
	 * @param r Recombination rate
	 * @return Fisher's amount of information
	 */
	public double getFisherInfo_F2_CD(double n, double r) {
		if(r == 0) return Double.MAX_VALUE;
		return 0.25 * n * ((3 + r) / (1 - r) + (1 - r) / (1 + r) + (4 - r) / r + r / (2 - r) + 2 * (2 * r * r - 2 * r - 1) / (1 - r + r * r));
	}

	/**
	 * Calculate standard deviation of recombination rate between codominant and dominant markers in F2 generation
	 * @param n # of individuals
	 * @param r Recombination rate
	 * @return ST error of recombinatino rate
	 */
	public double getStd_F2_CD(double n, double r) {
		return 1 / Math.sqrt(getFisherInfo_F2_CD(n, r));
	}

	/**
	 * Calculate Fisher's amount of information between domiant markers in coupling phase in F2 generation
	 * @param n # of individuals
	 * @param r Recombination rate
	 * @return Fisher's amount of information
	 */
	public double getFisherInfo_F2_DD_Cup(double n, double r) {
		if(r == 0) return Double.MAX_VALUE;
		return 2 * n * (3 - 4 * r + 2 * r * r) / (r * (2 - r) * (3 - 2 * r + r * r));
	}

	/**
	 * Calculate standard deviation of recombination rate between dominant markers in couplihng phase in F2 generation
	 * @param n # of individuals
	 * @param r Recombination rate
	 * @return ST error of recombinatino rate
	 */
	public double getStd_F2_DD_Cup(double n, double r) {
		return 1 / Math.sqrt(getFisherInfo_F2_DD_Cup(n, r));
	}

	/**
	 * Calculate Fisher's amount of information between domiant markers in repulsion phase in F2 generation
	 * @param n # of individuals
	 * @param r Recombination rate
	 * @return Fisher's amount of information
	 */
	public double getFisherInfo_F2_DD_Rep(double n, double r) {
		return 2 * n * (1 + 2 * r * r) / ((1 - r * r) * (2 + r * r));
	}

	/**
	 * Calculate standard deviation of recombination rate between dominant markers in repulsion phase in F2 generation
	 * @param n # of individuals
	 * @param r Recombination rate
	 * @return ST error of recombinatino rate
	 */
	public double getStd_F2_DD_Rep(double n, double r) {
		return 1 / Math.sqrt(getFisherInfo_F2_DD_Rep(n, r));
	}

	/**
	 * Calculate Fisher's amount of information in BC generation
	 * @param n # of individuals
	 * @param r Recombination rate
	 * @return Fisher's amount of information
	 */
	public double getFisherInfo_BC(double n, double r) {
		if(r == 0) return Double.MAX_VALUE;
		return n / (r * (1 - r));
	}

	/**
	 * Calculate standard deviation of recombination rate in BC generation
	 * @param n # of individuals
	 * @param r Recombination rate
	 * @return ST error of recombinatino rate
	 */
	public double getStd_BC(double n, double r) {
		return 1 / Math.sqrt(getFisherInfo_BC(n, r));
	}

	/**
	 * Calculate Fisher's amount of information in DH generation
	 * @param n # of individuals
	 * @param r Recombination rate
	 * @return Fisher's amount of information
	 */
	public double getFisherInfo_DH(double n, double r) {
		return getFisherInfo_BC(n, r);
	}

	/**
	 * Calculate standard deviation of recombination rate in DH generation
	 * @param n # of individuals
	 * @param r Recombination rate
	 * @return ST error of recombinatino rate
	 */
	public double getStd_DH(double n, double r) {
		return getStd_BC(n, r);
	}

	/**
	 * Calculate Fisher's amount of information in RIL generation
	 * @param n # of individuals
	 * @param r Recombination rate
	 * @return Fisher's amount of information
	 */
	public double getFisherInfo_RIL(double n, double r) {
		if(r == 0) return Double.MAX_VALUE;
		return 2 * n / (r * (1 + 2 * r) * (1 + 2 * r));
	}

	/**
	 * Calculate standard deviation of recombination rate in RIL generation
	 * @param n # of individuals
	 * @param r Recombination rate
	 * @return ST error of recombinatino rate
	 */
	public double getStd_RIL(double n, double r) {
		return 1 / Math.sqrt(getFisherInfo_RIL(n, r));
	}


}
