/*
AntMap: Software for constructing genetic linkage maps using an Ant Colony Optimization (ACO) algorithm
Copyright (C) 2004 National Agricultural and Bio-oriented Research Organization (NARO)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

Hiroyoshi IWATA
Department of Information Science and Technology
National Agricultural Research Center
3-1-1 Kannondai
Tsukuba
Ibaraki 305-8666
JAPAN

ph        +81-298-38-7025
fax       +81-298-38-8551
e-mail    iwatah@affrc.go.jp
*/

package org.agmodel.bioinfo.segregation;

import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;
import org.agmodel.bioinfo.common.StringFormat;

/**
 * @author iwata
 *
 */
public class SegDataFactory {
	private List chromoArray;
	private int nChromo;
	private int nMarker;

	public SegDataFactory() {
		chromoArray = new ArrayList();
		nChromo = 0;
		nMarker = 0;
	}

	public void addChromosome(Chromosome chr) {
		chromoArray.add(chr);
		nChromo ++;
		nMarker += chr.getNMarker();
	}

	public List getSegData(int nProgeny, int segType, double missingRate) {
		List segData = new ArrayList();

		segData.add("data type " + SegType.getSegTypeCode(segType)); // The first line of header
		segData.add(nProgeny + " " + nMarker + " " + " 0"); // The second line of header
		segData.add("");

		Iterator iterator = chromoArray.iterator();
		while(iterator.hasNext()) {
			Chromosome chromo = (Chromosome)iterator.next();
			char [][] segDataMatrix = getSegDataMatrix(nProgeny, chromo, segType, missingRate);
			String [] markerName = chromo.getMarkerNameArray();
			for(int i = 0; i < chromo.getNMarker(); i++ ) {
				String buf = new StringFormat().format("*"+ markerName[i], 10, true);
				for(int j = 0; j < nProgeny; j++ ) {
					if(j > 0 && j % 60 == 0)
						buf += "\n          ";
					buf += segDataMatrix[i][j];
				}
				segData.add(buf);
			}
		}

		return segData;
	}


	private char [][] getSegDataMatrix(int nProgeny, Chromosome chromo, int segType, double missingRate) {
		char [][] segDataMatrix = new char [chromo.getNMarker()] [nProgeny];
		for(int i = 0; i < nProgeny; i++ ) {
			int [] tmp = chromo.getPhenotype(segType, missingRate);
			for(int j = 0; j < tmp.length; j++ ) {
				char code;
				switch(tmp[j]) {
					case 1:
						code = 'A';
						break;
					case 2:
						code = 'B';
						break;
					case 3:
						code = 'H';
						break;
					case 4:
						code = 'C';
						break;
					case 5:
						code = 'D';
						break;
					default:
						code = '-';
				}
				segDataMatrix[j][i] = code;
			}
		}
		return segDataMatrix;
	}


}
