/*
AntMap: Software for constructing genetic linkage maps using an Ant Colony Optimization (ACO) algorithm
Copyright (C) 2004 National Agricultural and Bio-oriented Research Organization (NARO)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

Hiroyoshi IWATA
Department of Information Science and Technology
National Agricultural Research Center
3-1-1 Kannondai
Tsukuba
Ibaraki 305-8666
JAPAN

ph        +81-298-38-7025
fax       +81-298-38-8551
e-mail    iwatah@affrc.go.jp
*/

package org.agmodel.bioinfo.segregation;

import org.agmodel.bioinfo.stat.test.GOFChiSqrTest;

/**
 * @author iwata
 *
 */
public class SegRatioTest {
	private int segType = SegType.ST_NA;
	private int markerType = MarkerType.MT_NA;
	private int n_data = 0;
	private int n_miss = 0;
	private int n_class = 1;
	private char [] genotype = new char [n_class];
	private int [] n_obs = new int [n_class];
	private double [] n_exp = new double[n_class];
	private double chiSqr = 0.0;
	private double p = 0.0;

	/**
	 * test under normal segregation in segType generation
	 * @param segType segregation type
	 * @param l Array of genotypes
	 * @return result
	 */
	public SegRatioBean test(int segType, char [] l) {
		SegRatioBean segRatioBean;
		if(segType == SegType.ST_F2) {
			segRatioBean = testF2(l);
		} else if(segType == SegType.ST_BC) {
			segRatioBean = testBC(l);
		} else if(segType == SegType.ST_DH) {
			segRatioBean = testDH(l);
		} else if(segType == SegType.ST_RIL) {
			segRatioBean = testRIL(l);
		} else {
			segRatioBean = new SegRatioBean();
		}
		return segRatioBean;
	}


	/**
	 * test under normal segregation in BC generation
	 * @param l Array of genotypes for locus
	 * ('A' (homozygote for the recurrent parent genotype),
	 *  'H' (heterozogote),
	 *  '-' (missing data))
	 * @return result
	 */
	public SegRatioBean testBC(char [] l) {
		segType = SegType.ST_BC;
		markerType = MarkerType.MT_NA;
		n_data = l.length;
		n_miss = 0;
		n_class = 2;
		genotype = new char [n_class];
		genotype[0] = 'A';
		genotype[1] = 'H';
		n_exp = new double [n_class];
		n_exp[0] = (n_data - n_miss) * 0.5;
		n_exp[1] = (n_data - n_miss) * 0.5;
		n_obs = new int [n_class];
		for(int i = 0; i < l.length; i++ ) {
			if(l[i] == genotype[0]) {
				n_obs[0] ++;
			} else if(l[i] == genotype[1]) {
				n_obs[1] ++;
			} else {
				n_miss ++;
			}
		}
		GOFChiSqrTest gof = new GOFChiSqrTest(n_obs, n_exp);
		chiSqr = gof.getChiSqr();
		p = gof.getP();
		return setToSegRatioBean();
	}

	/**
	 * test under normal segregation in DH generation
	 * @param l Array of genotypes for locus
	 * ('A' (homozygote for parental genotype a),
	 *  'B' (homozygote for paretnal genotype b),
	 *  '-' (missing data))
	 * @return result
	 */
	public SegRatioBean testDH(char [] l) {
		segType = SegType.ST_DH;
		markerType = MarkerType.MT_NA;
		n_data = l.length;
		n_miss = 0;
		n_class = 2;
		genotype = new char [n_class];
		genotype[0] = 'A';
		genotype[1] = 'B';
		n_exp = new double [n_class];
		n_exp[0] = (n_data - n_miss) * 0.5;
		n_exp[1] = (n_data - n_miss) * 0.5;
		n_obs = new int [n_class];
		for(int i = 0; i < l.length; i++ ) {
			if(l[i] == genotype[0]) {
				n_obs[0] ++;
			} else if(l[i] == genotype[1]) {
				n_obs[1] ++;
			} else {
				n_miss ++;
			}
		}
		GOFChiSqrTest gof = new GOFChiSqrTest(n_obs, n_exp);
		chiSqr = gof.getChiSqr();
		p = gof.getP();
		return setToSegRatioBean();
	}

	/**
	 * get probability under normal segregation in RIL infinite generation
	 * @param l Array of genotypes for locus
	 * ('A' (homozygote for parental genotype a),
	 *  'B' (homozygote for paretnal genotype b),
	 *  '-' (missing data))
	 * @return result
	 */
	public SegRatioBean testRIL(char [] l) {
		segType = SegType.ST_RIL;
		markerType = MarkerType.MT_NA;
		n_data = l.length;
		n_miss = 0;
		n_class = 2;
		genotype = new char [n_class];
		genotype[0] = 'A';
		genotype[1] = 'B';
		n_exp = new double [n_class];
		n_exp[0] = (n_data - n_miss) * 0.5;
		n_exp[1] = (n_data - n_miss) * 0.5;
		n_obs = new int [n_class];
		for(int i = 0; i < l.length; i++ ) {
			if(l[i] == genotype[0]) {
				n_obs[0] ++;
			} else if(l[i] == genotype[1]) {
				n_obs[1] ++;
			} else {
				n_miss ++;
			}
		}
		GOFChiSqrTest gof = new GOFChiSqrTest(n_obs, n_exp);
		chiSqr = gof.getChiSqr();
		p = gof.getP();
		return setToSegRatioBean();
	}

	/**
	 * get probability under normal segregation in F2 generation
	 * @param l Array of genotypes for locus
	 * ('A' (homozygote for parental genotype a),
	 *  'B' (homozygote for paretnal genotype b),
	 *  'H' (heterozygote carrying both alleles a and b),
	 *  'C' (not a homozygote for allele a (either bb or ab genotype)),
	 *  'D' (not a homozygote for allele b (either aa or ab genotype)),
	 *  '-' (missing data))
	 * @return result
	 */
	public SegRatioBean testF2(char [] l) {
		segType = SegType.ST_RIL;
		markerType = new MarkerTypeGetter().getMarkerType(l);
		n_data = l.length;
		n_miss = 0;
		if(markerType == MarkerType.MT_CODOM) {
			n_class = 3;
			genotype = new char [n_class];
			genotype[0] = 'A';
			genotype[1] = 'H';
			genotype[2] = 'B';
			n_exp = new double [n_class];
			n_exp[0] = (n_data - n_miss) * 0.25;
			n_exp[1] = (n_data - n_miss) * 0.5;
			n_exp[2] = (n_data - n_miss) * 0.25;
			n_obs = new int [n_class];
			for(int i = 0; i < l.length; i++ ) {
				if(l[i] == genotype[0]) {
					n_obs[0] ++;
				} else if(l[i] == genotype[1]) {
					n_obs[1] ++;
				} else if(l[i] == genotype[2]) {
					n_obs[2] ++;
				} else {
					n_miss ++;
				}
			}
		} else if(markerType == MarkerType.MT_DOM_P1) {
			n_class = 2;
			genotype = new char [n_class];
			genotype[0] = 'D';
			genotype[1] = 'B';
			n_exp = new double [n_class];
			n_exp[0] = (n_data - n_miss) * 0.75;
			n_exp[1] = (n_data - n_miss) * 0.25;
			n_obs = new int [n_class];
			for(int i = 0; i < l.length; i++ ) {
				if(l[i] == genotype[0]) {
					n_obs[0] ++;
				} else if(l[i] == genotype[1]) {
					n_obs[1] ++;
				} else {
					n_miss ++;
				}
			}
		} else {
			n_class = 2;
			genotype = new char [n_class];
			genotype[0] = 'C';
			genotype[1] = 'A';
			n_exp = new double [n_class];
			n_exp[0] = (n_data - n_miss) * 0.75;
			n_exp[1] = (n_data - n_miss) * 0.25;
			n_obs = new int [n_class];
			for(int i = 0; i < l.length; i++ ) {
				if(l[i] == genotype[0]) {
					n_obs[0] ++;
				} else if(l[i] == genotype[1]) {
					n_obs[1] ++;
				} else {
					n_miss ++;
				}
			}
		}

		GOFChiSqrTest gof = new GOFChiSqrTest(n_obs, n_exp);
		chiSqr = gof.getChiSqr();
		p = gof.getP();
		return setToSegRatioBean();
	}

	private SegRatioBean setToSegRatioBean() {
		SegRatioBean segRatioBean = new SegRatioBean();
		segRatioBean.setN_data(n_data);
		segRatioBean.setN_miss(n_miss);
		segRatioBean.setGenotype(genotype);
		segRatioBean.setN_obs(n_obs);
		segRatioBean.setN_exp(n_exp);
		segRatioBean.setChiSqr(chiSqr);
		segRatioBean.setP(p);
		return segRatioBean;
	}
}
